// program to convert a KnotPlot floating point depth map image into
// a Portable Grey Map image

// 14 Nov 2005, Rob Scharein


#include <stdio.h>
#include <stdlib.h>

#define KNOTPLOT_DEPTH_MAP_MAGIC_NUMBER 1263551565 // the string `KPDM' as an int

#define MIN(X, Y)  ((X) < (Y) ? (X) : (Y))
#define MAX(X, Y)  ((X) > (Y) ? (X) : (Y))
#define CLAMP(V,L,H)   V = MAX (L, MIN (V, H))

// KnotPlot always uses network byte order for binary files.
// On IRIX and MacOSX this is the native byte order.
// Windows and most flavours of Linux use the opposite byte order. 

void read_int (int *v, FILE *fp) {     // read an int in network byte order
  unsigned char uc;
  *v = 0;
  for (int shift = 24; shift >= 0; shift -= 8) {
    fread (&uc, 1, 1, fp);
    *v |= uc << shift;
  }
}

void read_float (float *v, FILE *fp) { // read a float in network byte order
  union {
    int lg;
    float flt;
  } x;

  read_int (&x.lg, fp);
  *v = x.flt;
}

/*

  the KnotPlot commands 

  imgout depth gumby
  imgout depth pokey.fdm

  followed by 

  fdm2pgm pokey.fdm pokey.pgm

  on the command line should yield two files, `gumby.pgm' and `pokey.pgm' that are byte for byte identical

  */

int main (int argc, char **argv) {
  int magic, xsize, ysize;
  float near_value, far_value, depth_near, depth_far;
  FILE *fp_in, *fp_out;

  if (argc < 3) {
    fprintf (stderr, "Usage: %s fdmFile pgmFile\n", argv [0]);
    fprintf (stderr, "       where fdmFile is a KnotPlot floating point depth map file\n");
    fprintf (stderr, "       and pgmFile is the name of a PGM file for output.\n");
    exit (102);
  }
  
  fp_in = fopen (argv [1], "rb");   // need to open in binary (default in Windows is text)
  if (!fp_in) {
    fprintf (stderr, "  *** Can't open `%s'.\n", argv [1]);
    exit (1022);
  }

  read_int (&magic, fp_in);

  if (magic != KNOTPLOT_DEPTH_MAP_MAGIC_NUMBER) {
    fprintf (stderr, "  *** File `%s' is not a KnotPlot depth map!\n",
      argv [1]);
    exit (1022);
  }

  read_int (&xsize, fp_in);
  read_int (&ysize, fp_in);
  read_float (&near_value, fp_in);    // ignore these values for this program
  read_float (&far_value, fp_in);
  read_float (&depth_near, fp_in);
  read_float (&depth_far, fp_in);

  
  fp_out = fopen (argv [2], "wb");  // need to open in binary (default in Windows is text)
  if (!fp_out) {
    fprintf (stderr, "  *** Can't open `%s' for output.\n", argv [2]);
    exit (1022);
  }
  
  fprintf (fp_out, "P5\n");                       // identifies the file as a Portable Grey Map (PGM)
  fprintf (fp_out, "%d %d\n255\n", xsize, ysize); // image dimensions and maximum grey value
    
  for (int r = ysize - 1; r >= 0; r--) {      // top to bottom
    for (int c = 0; c < xsize; c++) {         // left to right
      unsigned char ucvalue;
      float depth;

      read_float (&depth, fp_in);              // note that values of depth can range outside the interval [0, 1]
      CLAMP (depth, (float) 0.0, (float) 1.0); // so clamp the value to the interval [0, 1]
      ucvalue = (unsigned char) (255.0 * depth + 0.5);  // scale by maximum grey level
      fwrite (&ucvalue, 1, 1, fp_out);
    }
  }
  
  fclose (fp_in);
  fclose (fp_out);

  return 0;
}

